"""Everything used for downloading video in bilibili.com."""
from urllib.parse import urljoin
import asyncio
import json
import re

from video_dl.spider import Spider
from video_dl.toolbox import info
from video_dl.video import Video, Media


class BilibiliSpider(Spider):
    """spider for bilibili.com"""
    site = 'bilibili.com'
    home_url = 'https://www.bilibili.com'

    # re patterns to extract information from html source code
    re_initial_state = re.compile(r'__INITIAL_STATE__=(.*?);\(function\(\)')
    re_playinfo = re.compile(r'window.__playinfo__=(.*?)</script>')

    def __init__(self):
        super().__init__()

        # video definition's id to definition's description
        self.id2desc = None
        self.video_pages = set()

    async def before_download(self) -> None:
        # parse origin html
        await self.parse_html(self.url)

        if self.lists and self.video_pages:
            info('list', 'tring to fetch more videos...')
            await asyncio.wait([asyncio.create_task(self.parse_html(url))
                                for url in self.video_pages])

        # choose media resource
        for video in self.video_list:
            video.choose_collection()

    def after_downloaded(self) -> None:
        for video in self.video_list:
            video.merge()

    async def parse_html(self, target_url: str) -> None:
        """extract key information from html source code.

        Args:
            target_url: target url copied from online vide website.
        """
        info('url', target_url)

        video = Video(self.session)  # need a session to access internet

        resp = await self.fetch_html(target_url)

        state = json.loads(self.re_initial_state.search(resp).group(1))
        current_page = state['p']
        pages = state['videoData']['pages']
        for page in pages:
            if page['page'] == current_page:
                video.title = page['part']
            else:
                # if there is other video in the same playlist
                video.parent_folder = state['videoData']['title']  # set folder
                if len(pages) > len(self.video_pages) + 1:
                    self.video_pages.add(  # ready to create task
                        urljoin(self.url, f'?p={page["page"]}')
                    )

        playinfo = json.loads(self.re_playinfo.search(resp).group(1))
        if self.id2desc is None:
            desc = playinfo['data']['accept_description']
            quality = playinfo['data']['accept_quality']
            self.id2desc = {
                str(key): value for key, value in zip(quality, desc)
            }

        videos = playinfo['data']['dash']['video']
        for media in videos:
            video.add_media(Media(
                url=media['base_url'],
                size=media['bandwidth'],
                desc=self.id2desc[str(media['id'])] + ' + ' + media['codecs'],
            ), target='picture')

        audios = playinfo['data']['dash']['audio']
        for media in audios:
            video.add_media(Media(
                url=media['base_url'],
                size=media['bandwidth']), target='sound')

        self.video_list.append(video)



"""Base class of all Spiders.

Typical usage:
    url = 'https://www.bilibili.com/video/BV15L411p7M8'
    spider = Spider.create_spider(url)  # will return a BilibiliSpider object
    asycio.run(spider.run())  # try to fetch resource information and download
"""
from urllib.parse import urlparse
import aiohttp
import asyncio

from video_dl.args import Arguments
from video_dl.toolbox import UserAgent, info


class Spider(object):
    """crawl data from web and download video.

    subclass of Spider should provide two public attributes:
        site: this class will use this field to create a specific Spider for
            target url. for example, BilibiliSpider's site is 'bilibili.com',
            will auto match to a url like 'https://www.bilibili.com/video/*'.
        home_url: target website's home page. this field will be inserted into
            headers of session to avoid some `no referer, no download` policy.

    subclass of Spider should implement some public methods:
        before_download: do something before download, just like: parse html.
        after_download: merge picture and sound to a completed video, delete
            tamporary files, and et al..
    """
    arg = Arguments()

    cookie = arg.cookie
    diretory = arg.directory
    proxy = arg.proxy
    url = arg.url
    lists = arg.lists

    @classmethod
    def create_spider(cls, url: str):
        """create a specific subclass depends on url."""
        netloc = urlparse(url).netloc
        for subclass in cls.__subclasses__():
            if subclass.site in netloc:
                return subclass()
        raise NotImplementedError

    def __init__(self):
        self.session = None
        self.headers = {
            'accept': '*/*',
            'user-agent': UserAgent().random,
            'cookie': self.cookie,
            'referer': self.home_url,
            'origin': self.home_url,
        }

        # list that contains Videos
        self.video_list = []

    async def create_session(self) -> None:
        """create client seesion if not exist."""
        if self.session is None:
            self.session = aiohttp.ClientSession(headers=self.headers)

    async def close_session(self) -> None:
        """close client session if possible."""
        if self.session:
            await self.session.close()

    async def fetch_html(self, url: str) -> str:
        """get url's html source code from internet."""
        async with self.session.get(url=url, proxy=self.proxy) as r:
            return await r.text()

    async def before_download(self) -> None:
        """do something before download"""
        raise NotImplementedError

    async def downloading(self) -> None:
        """download video from web."""
        await asyncio.wait([
            asyncio.create_task(video.download()) for video in self.video_list
        ])

    def after_downloaded(self) -> None:
        """do something after downloaded video."""
        pass

    async def run(self) -> None:
        """start crawl."""
        info('site', self.site)
        await self.create_session()

        await self.before_download()
        await self.downloading()
        self.after_downloaded()

        await self.close_session()


# import subclasses of Spider.
# You should import XxxSpider in sites/__init__.py
# then Spider.__subclasses__() could be workable.
import video_dl.sites  # pylint: disable=import-error
